@@ -32,7 +32,7 @@ details. */
 
 typedef void *HPCON;
 
-#include <devices.h>
+#include "devices.h"
 class tty_min
 {
   pid_t sid;	/* Session ID of tty */
@@ -103,17 +103,17 @@ public:
   };
 
 private:
+  HANDLE _from_master_nat;
   HANDLE _from_master;
-  HANDLE _from_master_cyg;
+  HANDLE _to_master_nat;
   HANDLE _to_master;
-  HANDLE _to_master_cyg;
+  HANDLE _to_slave_nat;
   HANDLE _to_slave;
-  HANDLE _to_slave_cyg;
   bool pcon_activated;
   bool pcon_start;
   pid_t pcon_start_pid;
   bool switch_to_pcon_in;
-  pid_t pcon_pid;
+  DWORD pcon_pid;
   UINT term_code_page;
   DWORD pcon_last_time;
   HANDLE h_pcon_write_pipe;
@@ -132,20 +132,21 @@ private:
   xfer_dir pcon_input_state;
   bool mask_flusho;
   bool discard_input;
+  bool stop_fwd_thread;
 
 public:
+  HANDLE from_master_nat () const { return _from_master_nat; }
   HANDLE from_master () const { return _from_master; }
-  HANDLE from_master_cyg () const { return _from_master_cyg; }
+  HANDLE to_master_nat () const { return _to_master_nat; }
   HANDLE to_master () const { return _to_master; }
-  HANDLE to_master_cyg () const { return _to_master_cyg; }
+  HANDLE to_slave_nat () const { return _to_slave_nat; }
   HANDLE to_slave () const { return _to_slave; }
-  HANDLE to_slave_cyg () const { return _to_slave_cyg; }
+  void set_from_master_nat (HANDLE h) { _from_master_nat = h; }
   void set_from_master (HANDLE h) { _from_master = h; }
-  void set_from_master_cyg (HANDLE h) { _from_master_cyg = h; }
+  void set_to_master_nat (HANDLE h) { _to_master_nat = h; }
   void set_to_master (HANDLE h) { _to_master = h; }
-  void set_to_master_cyg (HANDLE h) { _to_master_cyg = h; }
+  void set_to_slave_nat (HANDLE h) { _to_slave_nat = h; }
   void set_to_slave (HANDLE h) { _to_slave = h; }
-  void set_to_slave_cyg (HANDLE h) { _to_slave_cyg = h; }
 
   int read_retval;
   bool was_opened;	/* True if opened at least once. */
